<?php

namespace Drupal\static_export_graphql_data_resolver;

use Drupal\Core\Session\AccountSwitcherInterface;
use Drupal\Core\Session\UserSession;
use Drupal\graphql\GraphQL\Execution\QueryProcessor;
use Drupal\static_suite\Entity\EntityUtils;
use Drupal\static_suite\StaticSuiteUserException;
use Drupal\static_suite\Utility\StaticSuiteUtils;
use GraphQL\Server\OperationParams;

/**
 * A handler for querying GraphQL.
 */
class GraphqlQueryHandler implements GraphqlQueryHandlerInterface {

  /**
   * The account switcher.
   *
   * @var \Drupal\Core\Session\AccountSwitcherInterface
   */
  protected $accountSwitcher;

  /**
   * GraphQL query processor.
   *
   * @var \Drupal\graphql\GraphQL\Execution\QueryProcessor
   */
  protected $queryProcessor;

  /**
   * Entity Utils.
   *
   * @var \Drupal\static_suite\Entity\EntityUtils
   */
  protected $entityUtils;

  /**
   * The static suite utils.
   *
   * @var \Drupal\static_suite\Utility\StaticSuiteUtils
   */
  protected $staticSuiteUtils;

  /**
   * Constructor.
   *
   * @param \Drupal\Core\Session\AccountSwitcherInterface $accountSwitcher
   *   The account switcher.
   * @param \Drupal\graphql\GraphQL\Execution\QueryProcessor $queryProcessor
   *   GraphQL query processor.
   * @param \Drupal\static_suite\Entity\EntityUtils $entityUtils
   *   Entity utils service.
   * @param \Drupal\static_suite\Utility\StaticSuiteUtils $staticSuiteUtils
   *   The static suite utils.
   */
  public function __construct(AccountSwitcherInterface $accountSwitcher, QueryProcessor $queryProcessor, EntityUtils $entityUtils, StaticSuiteUtils $staticSuiteUtils) {
    $this->accountSwitcher = $accountSwitcher;
    $this->queryProcessor = $queryProcessor;
    $this->entityUtils = $entityUtils;
    $this->staticSuiteUtils = $staticSuiteUtils;
  }

  /**
   * {@inheritdoc}
   */
  public function query(string $graphqlQuery, array $variables = []): array {
    // Ensure "language" format (a common variable needed for querying multilingual
    // sites) is correct.
    if (isset($variables['language'])) {
      $variables['language'] = strtoupper(str_replace('-', '_', $variables['language']));
    }

    $bodyParams = [
      'query' => $graphqlQuery,
      'variables' => json_encode($variables),
      'operationName' => NULL,
    ];

    $params = OperationParams::create($bodyParams, TRUE);
    // When running on CLI, there is no user context and GraphQL queries are
    // executed as an anonymous user, leading to several problems and
    // inconsistencies: data that is properly resolved when logged on Drupal's
    // admin becomes non queryable when executing the same query on CLI.
    if ($this->staticSuiteUtils->isRunningOnCli()) {
      $this->accountSwitcher->switchTo(new UserSession(['uid' => 1]));
    }
    $result = $this->queryProcessor->processQuery('default:default', $params);
    if ($this->staticSuiteUtils->isRunningOnCli()) {
      $this->accountSwitcher->switchBack();
    }

    if (!empty($result->errors) && is_array($result->errors)) {
      throw new StaticSuiteUserException('Error on query: ' . $result->errors[0]);
    }

    return $result->toArray();
  }

}
