<?php

namespace Drupal\static_export_graphql_data_resolver\Plugin\static_export\Data\Resolver;

use Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException;
use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\static_export\Exporter\Data\Resolver\DataResolverPluginBase;
use Drupal\static_export\Exporter\ExporterPluginInterface;
use Drupal\static_export_graphql_data_resolver\GraphqlQueryHandlerInterface;
use Drupal\static_suite\Entity\EntityUtils;
use Drupal\static_suite\StaticSuiteUserException;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Serializer\NameConverter\CamelCaseToSnakeCaseNameConverter;

/**
 * A GraphQL resolver.
 *
 * @StaticDataResolver(
 *  id = "graphql",
 *  label = "GraphQL",
 *  description = "GraphQL data resolver",
 * )
 */
class GraphqlDataResolver extends DataResolverPluginBase {

  public const GRAPHQL_EXTENSION = '.gql';

  /**
   * Config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * Language Manager.
   *
   * @var \Drupal\Core\Language\LanguageManagerInterface
   */
  protected $languageManager;

  /**
   * Entity Utils.
   *
   * @var \Drupal\static_suite\Entity\EntityUtils
   */
  protected $entityUtils;

  /**
   * Handler for querying GraphQL.
   *
   * @var \Drupal\static_export_graphql_data_resolver\GraphqlQueryHandlerInterface
   */
  protected $graphqlQueryHandler;

  /**
   * Constructor.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   Config factory.
   * @param \Drupal\Core\Language\LanguageManagerInterface $languageManager
   *   Language Manager.
   * @param \Drupal\static_suite\Entity\EntityUtils $entityUtils
   *   Entity utils service.
   * @param \Drupal\static_export_graphql_data_resolver\GraphqlQueryHandlerInterface $graphqlQueryHandler
   *   Service for querying GraphQL.
   */
  public function __construct(array $configuration,
  string $plugin_id,
  $plugin_definition,
  ConfigFactoryInterface $config_factory,
  LanguageManagerInterface $languageManager,
  EntityUtils $entityUtils,
  GraphqlQueryHandlerInterface $graphqlQueryHandler) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->configFactory = $config_factory;
    $this->languageManager = $languageManager;
    $this->entityUtils = $entityUtils;
    $this->graphqlQueryHandler = $graphqlQueryHandler;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get("config.factory"),
      $container->get("language_manager"),
      $container->get("static_suite.entity_utils"),
      $container->get("static_export_graphql_data_resolver.graphql_query_handler")
    );
  }

  /**
   * {@inheritdoc}
   *
   * @throws \Drupal\static_suite\StaticSuiteUserException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  public function resolve(EntityInterface $entity, string $variant = NULL): array {
    $query = NULL;
    /** @var $exportableEntity \Drupal\static_export\Entity\ExportableEntity */
    $exportableEntity = $this->entityUtils->loadEntity('exportable_entity', $entity->getEntityTypeId() . '.' . $entity->bundle());
    if ($exportableEntity && $exportableEntity->status()) {
      // @todo - add event to alter the gql file path.
      $graphqlFile = $this->configFactory->get('static_export_graphql_data_resolver.settings')
        ->get('dir') . '/' . $exportableEntity->getEntityTypeIdString() . '/' . $exportableEntity->id() . ($variant ? ExporterPluginInterface::VARIANT_SEPARATOR . $variant : '') . self::GRAPHQL_EXTENSION;
      if (is_file($graphqlFile)) {
        // @todo - add event to alter the contents of $graphqlFile.
        $query = file_get_contents($graphqlFile);
      }
      else {
        throw new StaticSuiteUserException("Resolver query file located at '" . $graphqlFile . "' not found.");
      }
    }
    else {
      throw new StaticSuiteUserException("ExportableEntity for entity " . $entity->id() . " not found.");
    }

    // Any GraphQL query should be resolvable using these variables.
    $entityLanguage = $this->entityUtils->getNormalizedEntityLanguage($entity, FALSE);
    $entityLanguageId = $entityLanguage ? $entityLanguage->getId() : NULL;
    $variables = [
      'entityId' => $entity->id(),
      'uuid' => $entity->uuid(),
      'entityLanguageId' => strtoupper(str_replace('-', '_', $entityLanguageId)),
      'defaultLanguageId' => strtoupper(str_replace('-', '_', $this->languageManager->getDefaultLanguage()->getId())),
    ];
    // @todo - add event to alter $variables.
    // Throw an error if query is using nodeById without a "content" alias.
    $converter = new CamelCaseToSnakeCaseNameConverter();
    $byIdQueryName = $converter->denormalize($entity->bundle()) . 'ById';

    if (strpos($query, $byIdQueryName . '(') !== FALSE && !preg_match("/content:\s*$byIdQueryName\(/", $query)) {
      throw new StaticSuiteUserException("Resolver query file located at '" . $graphqlFile . "' is using @byIdQueryName() query without a 'content' alias. It should be aliased this way: 'content:@byIdQueryName(...) { ... }'", ['@byIdQueryName' => $byIdQueryName]);
    }

    return $this->graphqlQueryHandler->query($query, $variables);
  }

  /**
   * {@inheritdoc}
   */
  public function getVariantKeys(EntityInterface $entity): array {
    $variantKeys = [];
    /** @var $exportableEntity \Drupal\static_export\Entity\ExportableEntity */
    try {
      $exportableEntity = $this->entityUtils->loadEntity('exportable_entity', $entity->getEntityTypeId() . '.' . $entity->bundle());
    }
    catch (InvalidPluginDefinitionException | PluginNotFoundException $e) {
      return $variantKeys;
    }
    if ($exportableEntity && $exportableEntity->status()) {
      $basePath = $this->configFactory->get('static_export_graphql_data_resolver.settings')
        ->get('dir') . '/' . $exportableEntity->getEntityTypeIdString() . '/' . $exportableEntity->id() . ExporterPluginInterface::VARIANT_SEPARATOR;
      $pattern = $basePath . '*' . self::GRAPHQL_EXTENSION;
      $files = glob($pattern);
      foreach ($files as $file) {
        $variantFilename = str_replace($basePath, '', $file);
        $variantKey = pathinfo($variantFilename, PATHINFO_FILENAME);
        $variantKeys[] = $variantKey;
      }
    }
    return $variantKeys;
  }

}
