<?php

namespace Drupal\static_export\Plugin\static_export\Exporter\Config;

use Drupal\Component\Utility\NestedArray;
use Drupal\language\ConfigurableLanguageManagerInterface;
use Drupal\static_export\Exporter\Type\Config\ConfigExporterPluginBase;
use Drupal\static_export\File\FileCollection;
use Drupal\static_export\Exporter\ExporterPluginInterface;
use Drupal\static_export\Exporter\Output\Config\ExporterOutputConfigInterface;
use Drupal\static_suite\StaticSuiteUserException;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Exporter for Drupal configuration objects.
 *
 * @StaticConfigExporter(
 *  id = "default_config",
 *  label = @Translation("Configuration exporter"),
 *  description = @Translation("Exports configuration entities to filesystem. Default configuration exporter provided by Static Suite."),
 * )
 */
class ConfigExporter extends ConfigExporterPluginBase {

  public const PARAM_NAME = 'name';
  public const PARAM_LANG = 'language';
  protected const PARAM_INTERNAL_LANGCODE_BEING_DELETED = '__internal-langcode-being-deleted';

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The typed configuration manager.
   *
   * @var \Drupal\locale\LocaleConfigManager
   */
  protected $localeConfigManager;

  /**
   * The config exporter manager.
   *
   * @var \Drupal\static_export\Exporter\Type\Config\ConfigExporterPluginManagerInterface
   */
  protected $configExporterManager;

  /**
   * {@inheritdoc}
   */
  protected function setExtraDependencies(ContainerInterface $container): void {
    $this->moduleHandler = $container->get("module_handler");
    $this->localeConfigManager = $container->get("locale.config_manager");
    $this->configExporterManager = $container->get("plugin.manager.static_config_exporter");
  }

  /**
   * Array to hold several config objects.
   *
   * @var array
   */
  protected $exporterItem = [];

  /**
   * {@inheritdoc}
   *
   * @return \Drupal\Core\Config\Config
   *   Drupal configuration object.
   */
  public function getExporterItem() {
    if (empty($this->options[self::PARAM_NAME])) {
      return NULL;
    }

    if (is_array($this->exporterItem) && empty($this->exporterItem[$this->options[self::PARAM_NAME]])) {
      $this->exporterItem[$this->options[self::PARAM_NAME]] = $this->configFactory->get(
        $this->options[self::PARAM_NAME]
      );
    }

    return !empty($this->exporterItem[$this->options[self::PARAM_NAME]]) ? $this->exporterItem[$this->options[self::PARAM_NAME]] : NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function getExporterItemId() {
    return $this->getExporterItem()->getName();
  }

  /**
   * {@inheritdoc}
   */
  public function getExporterItemLabel() {
    return $this->getExporterItem()->getName();
  }

  /**
   * {@inheritdoc}
   */
  public function preProcessOptions(array $options): array {
    // Use default language if none defined.
    if (empty($options[self::PARAM_LANG])) {
      $options[self::PARAM_LANG] = $this->languageManager->getDefaultLanguage()->getId();
    }

    return $options;
  }

  /**
   * {@inheritdoc}
   *
   * @throws \Drupal\static_suite\StaticSuiteUserException
   */
  public function checkParams(array $options): bool {
    if (!isset($options[self::PARAM_NAME])) {
      throw new StaticSuiteUserException("Param '" . self::PARAM_NAME . "' is required.");
    }

    if (!$this->isMasterExport && isset($options[self::PARAM_INTERNAL_LANGCODE_BEING_DELETED])) {
      throw new StaticSuiteUserException("Param '" . self::PARAM_INTERNAL_LANGCODE_BEING_DELETED . "' is only allowed in a master export context.");
    }

    // Do not check if self::PARAM_NAME is exportable, since that is
    // done in $this->isExportable(), which is automatically called.
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  protected function getOutputDefinition(): ExporterOutputConfigInterface {
    $config = $this->configFactory->get('static_export.settings');
    $filename = $this->options[self::PARAM_NAME];
    $format = $config->get('exportable_config.format');

    // Load the OutputFormatter plugin definition to get its extension.
    $definitions = $this->outputFormatterManager->getDefinitions();
    $extension = !empty($definitions[$format]) ? $definitions[$format]['extension'] : $format;

    $outputConfig = $this->exporterOutputConfigFactory->create('', $filename, $extension);
    $outputConfig->getDefinition()->setFormat($format)->setLanguage($this->languageManager->getLanguage($this->options[self::PARAM_LANG]));
    return $outputConfig;
  }

  /**
   * {@inheritdoc}
   */
  protected function calculateDataFromResolver() {
    // Get config data with overrides applied.
    $originalData = $this->getExporterItem()->getOriginal('', TRUE);

    // Apply any translation override.
    if ($this->localeConfigManager->hasTranslation($this->options[self::PARAM_NAME], $this->options[self::PARAM_LANG]) && $this->languageManager instanceof ConfigurableLanguageManagerInterface) {
      $languageOverrideData = $this->languageManager->getLanguageConfigOverride($this->options[self::PARAM_LANG], $this->options[self::PARAM_NAME])->get();
      $originalData = NestedArray::mergeDeepArray([$originalData, $languageOverrideData], TRUE);
    }

    return $originalData;
  }

  /**
   * {@inheritdoc}
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\static_suite\StaticSuiteException
   * @throws \Drupal\static_suite\StaticSuiteUserException
   */
  public function exportTranslations(): FileCollection {
    $fileCollection = new FileCollection($this->uniqueId());
    if ($this->isMasterExport) {
      // Delete language if we receive __internal-langcode-being-deleted.
      if (!empty($this->options['__internal-langcode-being-deleted'])) {
        $fileCollection->mergeMultiple(
          $this->deleteSlave($this->options['__internal-langcode-being-deleted'])
        );
      }

      // Export languages.
      foreach ($this->getTranslationKeys() as $languageId) {
        if ($this->getOperation() === ExporterPluginInterface::OPERATION_WRITE) {
          $fileCollection->mergeMultiple(
            $this->writeSlave($languageId)
          );
        }
        else {
          $fileCollection->mergeMultiple(
            $this->deleteSlave($languageId)
          );
        }
      }

    }
    return $fileCollection;
  }

  /**
   * Helper method to trigger slave writes.
   *
   * @param string $language
   *   Language to be written.
   *
   * @return \Drupal\static_export\File\FileCollection[]
   *   An array of File Collections, ready to be merged with other ones.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\static_suite\StaticSuiteException
   * @throws \Drupal\static_suite\StaticSuiteUserException
   */
  protected function writeSlave(string $language): array {
    $slaveExporter = $this->configExporterManager->createDefaultInstance();
    $fileCollectionGroup = $slaveExporter->makeSlaveOf($this)
      ->write(
        [
          self::PARAM_NAME => $this->options[self::PARAM_NAME],
          self::PARAM_LANG => $language,
        ],
        TRUE,
        $this->mustLogToFile(),
        $this->isLock()
      );
    return $fileCollectionGroup->getFileCollections();
  }

  /**
   * Helper method to trigger slave deletes.
   *
   * @param string $language
   *   Language to be deleted.
   *
   * @return \Drupal\static_export\File\FileCollection[]
   *   An array of File Collections, ready to be merged with other ones.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\static_suite\StaticSuiteException
   * @throws \Drupal\static_suite\StaticSuiteUserException
   */
  protected function deleteSlave(string $language): array {
    $slaveExporter = $this->configExporterManager->createDefaultInstance();
    $fileCollectionGroup = $slaveExporter->makeSlaveOf($this)
      ->delete(
        [
          self::PARAM_NAME => $this->options[self::PARAM_NAME],
          self::PARAM_LANG => $language,
        ],
        TRUE,
        $this->mustLogToFile(),
        $this->isLock()
      );
    return $fileCollectionGroup->getFileCollections();
  }

  /**
   * Get exportable config names.
   *
   * @return array
   *   Array with exportable config names.
   */
  protected function getExportableConfigNames(): array {
    return $this->configFactory->get('static_export.settings')->get('exportable_config.objects_to_export');
  }

  /**
   * {@inheritdoc}
   *
   * @param string $item
   *   Config name.
   */
  public function isExportable(array $options): bool {
    $exportableConfigNames = $this->getExportableConfigNames();
    if (in_array($options[self::PARAM_NAME], $exportableConfigNames, TRUE)) {
      return TRUE;
    }
    return FALSE;
  }

  /**
   * {@inheritdoc}
   *
   * Configuration objects are somewhat special since they don't use
   * translations but overrides. That means that a configuration object is
   * always available in all languages: overrides are applied over the original
   * configuration object.
   *
   * To keep in sync with this strategy, configuration objects are always
   * exported in all languages, and overrides, if any, are applied to each of
   * them. This leads to data being repeated in several languages, but it makes
   * sense because avoids having to implement a more error-prone strategy (for
   * example, exporting only languages with overrides and then dynamically
   * loading the default configuration when no translated one is found).
   */
  protected function getTranslationKeyDefinitions(): array {
    $keys = [];
    foreach ($this->languageManager->getLanguages() as $language) {
      $languageId = $language->getId();
      // Skip the language that comes from options to avoid repeating it..
      if (($languageId !== $this->options[self::PARAM_LANG])/* && $this->localeConfigManager->hasTranslation($this->options[self::PARAM_NAME], $languageId)*/) {
        $keys[] = $languageId;
      }
    }
    return $keys;
  }

}
