<?php

namespace Drupal\static_suite\Entity;

use Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException;
use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Core\Entity\ContentEntityBase;
use Drupal\Core\Entity\EditorialContentEntityBase;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\Plugin\Validation\Constraint\BundleConstraint;
use Drupal\Core\Extension\ModuleExtensionList;
use Drupal\Core\Language\LanguageInterface;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\path_alias\AliasManagerInterface;
use Drupal\Core\Url;
use Throwable;

/**
 * A set of useful utils related to entities.
 */
class EntityUtils implements EntityUtilsInterface {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Language Manager.
   *
   * @var \Drupal\Core\Language\LanguageManagerInterface
   */
  protected $languageManager;

  /**
   * The alias manager.
   *
   * @var \Drupal\path_alias\AliasManagerInterface
   */
  protected $aliasManager;

  /**
   * Entity bundle id cache array.
   *
   * @var array
   */
  protected $entityBundleIdCache = [];

  /**
   * Constructs the entity utils.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager service.
   * @param \Drupal\Core\Language\LanguageManagerInterface $languageManager
   *   Language Manager.
   * @param \Drupal\path_alias\AliasManagerInterface $aliasManager
   *   The alias manager.
   * @param \Drupal\Core\Extension\ModuleExtensionList $extensionListModule
   *   List of available modules.
   */
  public function __construct(EntityTypeManagerInterface $entityTypeManager, LanguageManagerInterface $languageManager, AliasManagerInterface $aliasManager, ModuleExtensionList $extensionListModule) {
    $this->entityTypeManager = $entityTypeManager;
    $this->languageManager = $languageManager;
    $this->aliasManager = $aliasManager;
    $this->extensionListModule = $extensionListModule;
  }

  /**
   * {@inheritdoc}
   */
  public function loadEntity(string $entityTypeId, string $entityId, string $langcode = NULL): ?EntityInterface {
    try {
      $entity = $this->entityTypeManager->getStorage($entityTypeId)
        ->load($entityId);
    }
    catch (InvalidPluginDefinitionException | PluginNotFoundException $e) {
      $entity = NULL;
    }
    if ($entity && $langcode) {
      // All entities have a language even though they are not translatable.
      // First, find if entity language is the same they are asking for. This
      // works even for non translatable entities.
      $entityCurrentLanguage = $entity->language();
      if ($entityCurrentLanguage && $entityCurrentLanguage->getId() === $langcode) {
        return $entity;
      }
      // Second, find tranlations for translatable entities.
      if ($entity instanceof ContentEntityBase && $entity->isTranslatable()) {
        // Do not include the default language: it's been already processed.
        $translationLanguages = $entity->getTranslationLanguages(FALSE);
        if (isset($translationLanguages[$langcode])) {
          return $entity->getTranslation($langcode);
        }
      }
      return NULL;
    }
    return $entity;
  }

  /**
   * {@inheritdoc}
   */
  public function getEntityIds(string $entityTypeId, $options = []): array {
    $storage = $this->entityTypeManager->getStorage($entityTypeId);
    $query = $storage->getQuery();

    if (!empty($options['bundle'])) {
      $bundleEntityKey = $storage->getEntityType()->getKey('bundle');
      if (!empty($bundleEntityKey)) {
        $query->condition($bundleEntityKey, $options['bundle']);
      }
    }

    $hasStatus = $storage->getEntityType()->getKey('status');
    if ($hasStatus && !empty($options['status'])) {
      $query->condition('status', $options['status']);
    }

    if (isset($options['conditions'])) {
      foreach ($options['conditions'] as $condition) {
        if (!in_array($condition['field'], ['status', 'type'])) {
          $query->condition($condition['field'], $condition['value'], $condition['operator'] ?? NULL);
        }
      }
    }

    if (isset($options['sort'])) {
      $query->sort($options['sort']['field'], $options['sort']['direction']);
    }

    if (isset($options['range'])) {
      $query->range($options['range']['start'], $options['range']['length']);
    }

    return $query->execute();
  }

  /**
   * {@inheritdoc}
   */
  public function getEntityBundleId(EntityInterface $entity): ?string {
    if (!isset($this->entityBundleIdCache[$entity->id()])) {
      $entityBundleId = NULL;
      $constrains = $entity->getTypedData()->getConstraints();
      foreach ($constrains as $constrain) {
        if ($constrain instanceof BundleConstraint) {
          if (is_array($constrain->bundle)) {
            $entityBundleId = $constrain->bundle[0];
          }
          else {
            $entityBundleId = $constrain->bundle;
          }
          break;
        }
      }
      $this->entityBundleIdCache[$entity->id()] = $entityBundleId;
    }
    return $this->entityBundleIdCache[$entity->id()];
  }

  /**
   * {@inheritdoc}
   */
  public function hasStatusChanged(EditorialContentEntityBase $entity): bool {
    $statusHasChanged = FALSE;
    $entityOldStatus = $entity->isPublished();
    $originalEntity = $entity->original;
    if (is_object($originalEntity) && $originalEntity instanceof EditorialContentEntityBase) {
      $entityOldStatus = $originalEntity->isPublished();
    }
    $entityNewStatus = $entity->isPublished();
    if ($entityOldStatus !== $entityNewStatus) {
      $statusHasChanged = TRUE;
    }
    return $statusHasChanged;
  }

  /**
   * {@inheritdoc}
   */
  public function getEntityDataByPagePath(string $pagePath, string $langcode = NULL): ?array {
    $entityData = NULL;
    // $path can arrive without a leading slash, but
    // Drupal\Core\Path\AliasManager needs that leading slash.
    if ($pagePath[0] !== "/") {
      $pagePath = "/" . $pagePath;
    }
    try {
      $pagePath = $this->aliasManager->getPathByAlias($pagePath, $langcode);
      $params = Url::fromUri("internal:" . $pagePath)->getRouteParameters();
      $entityTypeId = key($params);
      $availableEntityTypes = $this->entityTypeManager->getDefinitions();
      if ($entityTypeId && isset($availableEntityTypes[$entityTypeId])) {
        $entityId = $params[$entityTypeId];
        if ($entityId) {
          $entityData = [
            'entityTypeId' => $entityTypeId,
            'entityId' => $entityId,
          ];
        }
      }
    }
    catch (Throwable $e) {
      // Do nothing.
    }
    return $entityData;
  }

  /**
   * {@inheritdoc}
   */
  public function getNormalizedEntityLanguage(EntityInterface $entity, bool $useLockedLanguages = TRUE): ?LanguageInterface {
    // Configurable languages are the normal ones (en, es, fr, etc)
    $configurableLanguages = $this->languageManager->getLanguages($useLockedLanguages ? LanguageInterface::STATE_ALL : LanguageInterface::STATE_CONFIGURABLE);
    $entityLangcode = $entity->language()->getId();
    if (isset($configurableLanguages[$entityLangcode])) {
      // Entity's language is among configurable languages, so use it.
      return $entity->language();
    }

    // Locked languages are LanguageInterface::LANGCODE_NOT_SPECIFIED and
    // LanguageInterface::LANGCODE_NOT_APPLICABLE (und and zxx).
    if (!$useLockedLanguages) {
      $lockedLanguages = $this->languageManager->getLanguages(LanguageInterface::STATE_LOCKED);
      if (isset($lockedLanguages[$entityLangcode])) {
        // Entity's language is not valid, so don't use it.
        return NULL;
      }
    }

    // Entity's language is something else (LanguageInterface::LANGCODE_SYSTEM,
    // LanguageInterface::LANGCODE_DEFAULT,
    // LanguageInterface::LANGCODE_SITE_DEFAULT, etc), so use the default one.
    return $this->languageManager->getDefaultLanguage();
  }

}
